#pragma once

#include <osmocom/core/utils.h>

extern const struct value_string octphy_l1sapi_names[23];
extern const struct value_string octphy_dir_names[5];
extern const struct value_string octphy_clkmgr_state_vals[8];
extern const struct value_string octphy_cid_vals[37];
extern const struct value_string octphy_eid_vals[7];
