#ifndef _OC2GBTS_TEMP_H
#define _OC2GBTS_TEMP_H

enum oc2gbts_temp_sensor {
	OC2GBTS_TEMP_SUPPLY,
	OC2GBTS_TEMP_SOC,
	OC2GBTS_TEMP_FPGA,
	OC2GBTS_TEMP_RMSDET,
	OC2GBTS_TEMP_OCXO,
	OC2GBTS_TEMP_TX,
	OC2GBTS_TEMP_PA,
	_NUM_TEMP_SENSORS
};

enum oc2gbts_temp_type {
	OC2GBTS_TEMP_INPUT,
	OC2GBTS_TEMP_LOWEST,
	OC2GBTS_TEMP_HIGHEST,
	OC2GBTS_TEMP_FAULT,
	_NUM_TEMP_TYPES
};

int oc2gbts_temp_get(enum oc2gbts_temp_sensor sensor, int *temp);


#endif
